@@ -3,7 +3,7 @@ package YAHC;
 use strict;
 use warnings;
 
-our $VERSION = '0.001';
+our $VERSION = '0.002';
 
 use EV;
 use Time::HiRes;
@@ -11,7 +11,7 @@ use Exporter 'import';
 use Scalar::Util qw/weaken/;
 use Fcntl qw/F_GETFL F_SETFL O_NONBLOCK/;
 use POSIX qw/EINPROGRESS EINTR EAGAIN EWOULDBLOCK strftime/;
-use Socket qw/PF_INET SOCK_STREAM $CRLF SOL_SOCKET SO_ERROR SO_RCVTIMEO SO_SNDTIMEO inet_aton pack_sockaddr_in/;
+use Socket qw/PF_INET SOCK_STREAM $CRLF SOL_SOCKET SO_ERROR SO_RCVTIMEO SO_SNDTIMEO inet_aton inet_ntoa pack_sockaddr_in/;
 
 sub YAHC::Error::NO_ERROR                () { 0 }
 sub YAHC::Error::REQUEST_TIMEOUT         () { 1 << 0 }
@@ -40,6 +40,8 @@ use constant {
     TCP_READ_CHUNK             => 65536,
     RUN_UNTIL_ALL_CONNECTED    => 'run until all connected',
     RUN_UNTIL_ALL_SENT_REQUEST => 'run until all sent request',
+    CALLBACKS                  => [ qw/init_callback wait_synack_callback connected_callback
+                                       writing_callback reading_callback completed_callback callback/ ],
 };
 
 our @EXPORT_OK = qw/
@@ -114,17 +116,15 @@ sub request {
     die "YAHC: Connection with name '$conn_id' already exists"
         if exists $self->{connections}{$conn_id};
 
-    my $cb = delete $request->{callback};
-    my $debug = delete $request->{debug};
-
     my $conn = {
         id          => $conn_id,
         request     => $request,
         response    => { status_code => 0 },
         attempt     => 0,
-        retries     => $request->{connection_retries} || 0,
+        retries     => $request->{retries} || 0,
         state       => YAHC::State::INITIALIZED(),
-        debug       => $debug || $self->{debug},
+        debug       => delete $request->{debug} || $self->{debug},
+        timeline    => delete $request->{timeline} || $self->{timeline},
         selected_target => [],
     };
 
@@ -133,8 +133,15 @@ sub request {
     do { $request->{$_} ||= $pool_args->{$_} if $pool_args->{$_} } foreach (qw/host port/);
     die 'YAHC: host must be defined' unless $request->{host};
 
+    my %callbacks;
+    foreach (@{ CALLBACKS() }) {
+        next unless $request->{$_};
+        $callbacks{$_} = delete $request->{$_};
+        $conn->{"has_$_"} = 1;
+    }
+
     $self->{watchers}{$conn_id} = {};
-    $self->{callbacks}{$conn_id} = $cb;
+    $self->{callbacks}{$conn_id} = \%callbacks;
     $self->{connections}{$conn_id} = $conn;
 
     $self->_set_request_timer($conn_id) if $conn->{request}{request_timeout};
@@ -254,8 +261,8 @@ sub _set_init_state {
 
     $conn->{response} = { status_code => 0 };
     $conn->{state} = YAHC::State::INITIALIZED();
-    _register_in_timeline($conn, "connection %s new state %s",
-                          $conn_id, _strstate($conn->{state})) if $conn->{debug};
+    _register_in_timeline($conn, "new state %s", _strstate($conn->{state})) if $conn->{timeline};
+    $self->_call_state_callback($conn, 'init_callback') if $conn->{has_init_callback};
 
     my $continue = 1;
     while ($continue) {
@@ -266,13 +273,13 @@ sub _set_init_state {
 
         my $attempt = $conn->{attempt}++;
         if ($attempt > $conn->{retries}) {
-            $self->_set_user_action_state($conn_id, YAHC::Error::CONNECT_ERROR(), "connection_retries limit reached");
+            $self->_set_user_action_state($conn_id, YAHC::Error::CONNECT_ERROR(), "retries limit reached");
             return;
         }
 
         eval {
             my ($host, $ip, $port) = _get_next_target($conn);
-            _register_in_timeline($conn, "Target $host:$port ($ip:$port) chosen for attempt #$attempt") if $conn->{debug};
+            _register_in_timeline($conn, "Target $host:$port ($ip:$port) chosen for attempt #$attempt") if $conn->{timeline};
 
             my $sock = _build_socket_and_connect($ip, $port, $conn->{request});
             $self->_set_wait_synack_state($conn_id, $sock, $ip, $host, $port);
@@ -297,12 +304,12 @@ sub _set_wait_synack_state {
     _assert_state($conn, YAHC::State::INITIALIZED()) if $conn->{debug};
 
     $conn->{state} = YAHC::State::WAIT_SYNACK();
-    _register_in_timeline($conn, "connection %s new state %s",
-                          $conn_id, _strstate($conn->{state})) if $conn->{debug};
+    _register_in_timeline($conn, "new state %s", _strstate($conn->{state})) if $conn->{timeline};
+    $self->_call_state_callback($conn, 'wait_synack_callback') if $conn->{has_wait_synack_callback};
 
     $self->_check_stop_condition($conn) if $self->{stop_condition};
 
-    my $wait_synack_cb = $self->_get_safe_wrapper($conn_id, sub {
+    my $wait_synack_cb = $self->_get_safe_wrapper($conn, sub {
         my $sockopt = getsockopt($sock, SOL_SOCKET, SO_ERROR);
         if (!$sockopt) {
             _register_error($conn, YAHC::Error::CONNECT_ERROR(), "Failed to do getsockopt(): $!");
@@ -318,8 +325,8 @@ sub _set_wait_synack_state {
         }
 
         $conn->{state} = YAHC::State::CONNECTED();
-        _register_in_timeline($conn, "connection %s new state %s",
-                              $conn_id, _strstate($conn->{state})) if $conn->{debug};
+        _register_in_timeline($conn, "new state %s", _strstate($conn->{state})) if $conn->{timeline};
+        $self->_call_state_callback($conn, 'connected_callback') if $conn->{has_connected_callback};
 
         $self->_set_write_state($conn_id);
     });
@@ -336,8 +343,8 @@ sub _set_write_state {
     _assert_connected($conn) if $conn->{debug};
 
     $conn->{state} = YAHC::State::WRITING();
-    _register_in_timeline($conn, "connection %s new state %s",
-                          $conn_id, _strstate($conn->{state})) if $conn->{debug};
+    _register_in_timeline($conn, "new state %s", _strstate($conn->{state})) if $conn->{timeline};
+    $self->_call_state_callback($conn, 'writing_callback') if $conn->{has_writing_callback};
 
     $self->_check_stop_condition($conn) if $self->{stop_condition};
 
@@ -345,9 +352,9 @@ sub _set_write_state {
     my $buf = _build_http_message($conn);
     my $length = length($buf);
 
-    _register_in_timeline($conn, "sending request of $length bytes") if $conn->{debug};
+    _register_in_timeline($conn, "sending request of $length bytes") if $conn->{timeline};
 
-    my $write_cb = $self->_get_safe_wrapper($conn_id, sub {
+    my $write_cb = $self->_get_safe_wrapper($conn, sub {
         my $wlen = POSIX::write($fd, $buf, $length);
 
         if (!$wlen) {
@@ -379,8 +386,8 @@ sub _set_read_state {
     _assert_connected($conn) if $conn->{debug};
 
     $conn->{state} = YAHC::State::READING();
-    _register_in_timeline($conn, "connection %s new state %s",
-                          $conn_id, _strstate($conn->{state})) if $conn->{debug};
+    _register_in_timeline($conn, "new state %s", _strstate($conn->{state})) if $conn->{debug};
+    $self->_call_state_callback($conn, 'reading_callback') if $conn->{has_reading_callback};
 
     $self->_check_stop_condition($conn) if $self->{stop_condition};
 
@@ -390,7 +397,7 @@ sub _set_read_state {
     my $content_length = 0;
     my $fd = fileno($watcher->fh);
 
-    my $read_cb = $self->_get_safe_wrapper($conn_id, sub {
+    my $read_cb = $self->_get_safe_wrapper($conn, sub {
         my $rlen = POSIX::read($fd, my $b = '', TCP_READ_CHUNK);
 
         if (!$rlen) {
@@ -417,7 +424,7 @@ sub _set_read_state {
                 $content_length = $headers->{'Content-Length'};
                 substr($buf, 0, 4, ''); # 4 = length("$CRLF$CRLF")
                 _register_in_timeline($conn, "headers parsed: content-length='%d' content-type='%s'",
-                                      $content_length, $headers->{'Content-Type'} || '<no-content-type>') if $conn->{debug};
+                                      $content_length, $headers->{'Content-Type'} || '<no-content-type>') if $conn->{timeline};
             }
 
             if ($decapitated && length($buf) >= $content_length) {
@@ -444,31 +451,26 @@ sub _set_user_action_state {
         return;
 
     $conn->{state} = YAHC::State::USER_ACTION();
-    _register_in_timeline($conn, "connection %s new state %s",
-                          $conn_id, _strstate($conn->{state})) if $conn->{debug};
+    _register_in_timeline($conn, "new state %s", _strstate($conn->{state})) if $conn->{timeline};
     _register_error($conn, $error, $strerror) if $error;
 
-    my $cb = $self->{callbacks}{$conn_id};
-    return $self->_set_completed_state($conn_id) unless $cb; # shortcut
+    return $self->_set_completed_state($conn_id) unless $conn->{has_callback};
+    my $cb = $self->{callbacks}{$conn_id}{callback};
 
     eval {
-        _register_in_timeline($conn, "call callback%s", $error ? " error=$error, strerror='$strerror'" : '') if $conn->{debug};
+        _register_in_timeline($conn, "call callback%s", $error ? " error=$error, strerror='$strerror'" : '') if $conn->{timeline};
         $cb->($conn, $error, $strerror);
         1;
     } or do {
         my $error = $@ || 'zombie error';
-        _register_error($conn, YAHC::Error::CALLBACK_ERROR(), $error);
+        _register_error($conn, YAHC::Error::CALLBACK_ERROR(), "Exception in callback: $error");
         warn "YAHC: exception in callback: $error";
-    };
-
-    if ($error == YAHC::Error::INTERNAL_ERROR()) {
-        _register_in_timeline($conn, "unrecoverable error appeared, forcing completion") if $conn->{debug};
         $self->_set_completed_state($conn_id);
         return;
-    }
+    };
 
     my $state = $conn->{state};
-    _register_in_timeline($conn, "after invoking callback state is %s", _strstate($state)) if $conn->{debug};
+    _register_in_timeline($conn, "after invoking callback state is %s", _strstate($state)) if $conn->{timeline};
 
     if ($state == YAHC::State::INITIALIZED()) {
         $self->_set_init_state($conn_id); # TODO eval
@@ -492,8 +494,18 @@ sub _set_completed_state {
         return;
 
     $conn->{state} = YAHC::State::COMPLETED();
-    _register_in_timeline($conn, "connection %s new state %s",
-                          $conn_id, _strstate($conn->{state})) if $conn->{debug};
+    _register_in_timeline($conn, "new state %s", _strstate($conn->{state})) if $conn->{timeline};
+
+    if ($conn->{has_completed_callback}) {
+        eval {
+            $self->{callbacks}{$conn_id}{completed_callback}->($conn);
+            1;
+        } or do {
+            my $error = $@ || 'zombie error';
+            _register_error($conn, YAHC::Error::CALLBACK_ERROR(), "Exception in callback: $error");
+            warn "YAHC: exception in callback: $error";
+        }
+    }
 
     if (my $w = delete $watchers->{io}) {
         $w->stop;
@@ -566,16 +578,16 @@ sub _set_request_timer {
     my $timeout = $conn->{request}{request_timeout};
     return unless defined $timeout;
 
-    my $request_timer_cb = $self->_get_safe_wrapper($conn_id, sub {
+    my $request_timer_cb = $self->_get_safe_wrapper($conn, sub {
         if ($conn->{state} < YAHC::State::USER_ACTION()) {
             my $error = sprintf("Request timeout of %.3fs has reached", $timeout);
             $self->_set_user_action_state($conn_id, YAHC::Error::REQUEST_TIMEOUT(), $error);
         } else {
-            _register_in_timeline($conn, "delete request timer") if $conn->{debug};
+            _register_in_timeline($conn, "delete request timer") if $conn->{timeline};
         }
     });
 
-    _register_in_timeline($conn, "setting request timeout to %.3fs", $timeout) if $conn->{debug};
+    _register_in_timeline($conn, "setting request timeout to %.3fs", $timeout) if $conn->{timeline};
     $watchers->{request_timer} = $self->{loop}->timer($timeout, 0, $request_timer_cb);
     $watchers->{request_timer}->priority(2); # set max priority
 }
@@ -594,16 +606,16 @@ sub _set_connection_timer {
     my $timeout = $conn->{request}{connect_timeout};
     return unless $timeout;
 
-    my $connection_timer_cb = $self->_get_safe_wrapper($conn_id, sub {
+    my $connection_timer_cb = $self->_get_safe_wrapper($conn, sub {
         if ($conn->{state} < YAHC::State::CONNECTED()) {
-            _register_in_timeline($conn, "connection timeout of %.3fs has reached", $timeout) if $conn->{debug};
+            _register_in_timeline($conn, "connection timeout of %.3fs has reached", $timeout) if $conn->{timeline};
             $self->_set_init_state($conn_id);
         } else {
-            _register_in_timeline($conn, "delete connection timer") if $conn->{debug};
+            _register_in_timeline($conn, "delete connection timer") if $conn->{timeline};
         }
     });
 
-    _register_in_timeline($conn, "setting connection timeout to %.3fs", $timeout) if $conn->{debug};
+    _register_in_timeline($conn, "setting connection timeout to %.3fs", $timeout) if $conn->{timeline};
     $watchers->{connection_timer} = $self->{loop}->timer($timeout, 0, $connection_timer_cb);
     $watchers->{connection_timer}->priority(-2); # set lowest priority
 }
@@ -670,14 +682,32 @@ sub _wrap_target_selection {
     die "YAHC: unsupported host format";
 }
 
+sub _call_state_callback {
+    my ($self, $conn, $cb_name) = @_;
+    my $cb = $self->{callbacks}{$conn->{id}}{$cb_name};
+    return unless $cb;
+
+    eval {
+        $cb->($conn);
+        1;
+    } or do {
+        my $error = $@ || 'zombie error';
+        _register_error($conn, YAHC::Error::CALLBACK_ERROR(), "Exception in callback: $error");
+        warn "YAHC: exception in callback: $error";
+        $self->_set_completed_state($conn->{id});
+    };
+}
+
 sub _get_safe_wrapper {
-    my ($self, $conn_id, $sub) = @_;
+    my ($self, $conn, $sub) = @_;
     return sub { eval {
         $sub->(@_);
         1;
     } or do {
         my $error = $@ || 'zombie error';
-        $self->_set_user_action_state($conn_id, YAHC::Error::INTERNAL_ERROR(), "Exception callback: $error");
+        _register_error($conn, YAHC::Error::INTERNAL_ERROR(), "Exception callback: $error");
+        warn "YAHC: exception in callback: $error";
+        $self->_set_completed_state($conn->{id});
     }};
 }
 
@@ -686,7 +716,7 @@ sub _register_in_timeline {
     my $event = sprintf("$format", @arguments);
     $event =~ s/\s+$//g;
     _log_message("YAHC connection '%s': %s", $conn->{id}, $event) if $conn->{debug};
-    push @{ $conn->{timeline} ||= [] }, [ $event, Time::HiRes::time ];
+    push @{ $conn->{timeline} ||= [] }, [ $event, $conn->{state}, Time::HiRes::time ] if $conn->{timeline};
 }
 
 sub _register_error {
